import collections
import collections.abc
import enum
import itertools
import random

CardSuit = enum.IntEnum('CardSuit', 'STICKS SWORDS CUPS COINS')
GameStatus = enum.IntEnum(
    'GameStatus', 'WAIT_PLAYERS WAIT_START PLAYING PAUSED ENDED')
HandStatus = enum.IntEnum('HandStatus', 'BET MOVE ENDED')


class InvalidOperation(Exception):
    """Tells that an operation cannot be done (at least at the moment)."""
    pass


class Card:
    """A class that represent a card (a number and a suit).

    It provides also some mechanics that are useful for the game, e.g.
    checking card equality, or order.
    Basically, providing these capabilities is the reason to have a
    class, instead of using e.g. tuples.

    The class handles also the highest/lowest card for the coin ace,
    which is set as flag on the class instance, but does not change
    the result of equality (i.e. a high coin ace and a low coin ace
    are equal, for this class), whereas of course it is handled
    correctly on the lesser/greater operators.
    XXX: should equality look at high?

    Moreover, having a class for cards makes it possible to implement
    variatons of the game based on other cards, e.g. French cards
    instead of Italian cards.
    """

    def __init__(self, number, suit, high=False):
        if not (1 <= number <= 10):
            raise ValueError('Invalid value for the card')
        if type(suit) != CardSuit:
            suit = CardSuit(suit)
        self.number = number
        self.suit = suit
        if self.eligible_high():
            self.high = high
        else:
            self.high = False

    def eligible_high(self):
        """Tells if a card can be set as highest/lowest card."""
        return (self.number == 1 and self.suit == CardSuit.COINS)

    def set_high(self, high=True):
        """Set the coin ace as high card."""
        if self.eligible_high():
            self.high = high
        else:
            raise InvalidOperation('Only the coin ace can be high card')

    def serialize(self):
        """Serialize a card as builtin types."""
        if self.eligible_high():
            return [self.number, self.suit.value, self.high]
        else:
            return [self.number, self.suit.value]

    def __str__(self):
        return '{} {}'.format(self.number, self.suit.name)

    def __repr__(self):
        # Use str(suit) instead of suit.name to have also the class
        # name and make repr as Python code
        return '{}({}, {}, {})'.format(
            type(self).__name__, self.number, str(self.suit), self.high)

    def __eq__(self, other):
        return self.number == other.number and self.suit == other.suit

    def __hash__(self):
        return self.suit * 100 + self.number

    def __gt__(self, other):
        if other is None:
            return True
        if self.eligible_high():
            return self.high and not other.high
        elif other.eligible_high():
            return not other.high
        return (self.suit > other.suit
                or (self.suit == other.suit and self.number > other.number))

    def __lt__(self, other):
        if other is None:
            # This allows to use None as initial max card value.
            # In this logic, None is lower even than coin ace used as
            # lowest card, so if lowest coin ace is the first card that
            # is moved, it also becomes the max card.
            return False
        if self.eligible_high():
            return not self.high and (not other.eligible_high() or other.high)
        elif other.eligible_high():
            return other.high
        return (self.suit < other.suit
                or (self.suit == other.suit and self.number < other.number))

    @classmethod
    def make_all(cls):
        """Return all possible cards."""
        return [cls(n, s) for n, s in itertools.product(
            range(1, 11), list(CardSuit))]

    @classmethod
    def make_cards(cls, num_players, num_cards):
        """Return some cards, grouped for a hand."""
        cards = cls.make_all()
        if num_players * num_cards > len(cards):
            raise ValueError('Not enough cards')
        random.shuffle(cards)
        return [sorted(cards[i:i + num_cards]) for i in range(
            0, num_players * num_cards, num_cards)]


class Player:
    """The default value for bet and took.
    Use the same value for bet and took to allow to reset even after
    construction or to reset multiple times.
    """
    DEFAULT = 0

    def __init__(self, lives, cards=None, bet=DEFAULT, took=DEFAULT):
        self.lives = lives
        self.cards = cards
        self.bet = bet
        self.took = took

    def reset(self, cards=None, bet=DEFAULT, took=DEFAULT):
        self.lives -= abs(self.bet - self.took)
        self.cards = cards
        self.bet = bet
        self.took = took

    def dead(self):
        """Tells if a player is dead."""
        return self.lives <= 0


class Game:
    """The class that handles the game mechanics.

    It has been designed to be as standalone as possible: it makes only
    the assumption that the class that manages card implements
    correctly the comparisons, and that it provides the highest/lowewst
    card mechanism.

    The Game class uses also a Player class, but only as an easier way
    to manage player data (an OrderedDict, instead of many lists), but
    it does not make assumptions on the object that are used as keys of
    the dictionary: they can be string, numbers or any object, provided
    that they are valid dictionary keys, that they are unique, and that
    the None value cannot be used to identify a player.

    Finally, this class has some callbacks, that can be used to make
    the game asynchronous, in a certain sense: when something important
    happens (e.g. a successful move, or a bet), a method is called,
    and an implementation can broadcast the event to all the players,
    instead of having to check whether the action succeeded, and in
    case broadcast the event.
    The indended way to use this feature is to extend this class.
    """

    """The default value for lives."""
    STARTING_LIVES = 5

    """The default value of cards in a full hand."""
    STARTING_CARDS = 5

    """This is a value based on 5 starting cards, and 40 card deck.
    In future it will be changed, to something better and less
    hardcoded.
    """
    MAX_PLAYERS = 8

    """Placeholder to allow variations of Card types."""
    CARD_CLASS = Card

    """Handle automatically obvious moves?"""
    ENABLE_AUTOMOVE = False

    def __init__(self, max_players=MAX_PLAYERS, starting_lives=STARTING_LIVES,
                 card_class=CARD_CLASS):
        # FIXME: Actually max_players depends on the starting cards.
        # Still, having it helps when joining players.
        # I might remove this in the future.
        if (not isinstance(max_players, int)
                or not (1 < max_players <= self.MAX_PLAYERS)):
            raise ValueError(
                'max_players must be an integer, between 2 and {}'.format(
                    self.MAX_PLAYERS))

        self.max_players = max_players
        self.starting_lives = starting_lives
        self.card_class = card_class
        self.enable_automove = self.ENABLE_AUTOMOVE
        self.players = collections.OrderedDict()
        self.dealer = None
        self.game_status = GameStatus.WAIT_PLAYERS
        self.hand_status = HandStatus.ENDED
        self.num_cards = 0

    def add_player(self, id_):
        """Add a player."""
        if self.game_status != GameStatus.WAIT_PLAYERS:
            raise InvalidOperation('Not accepting players at the moment')
        if id_ is None:
            raise ValueError('None cannot be used as player id')
        if id_ in self.players:
            raise ValueError('Player already registered')
        if len(self.players) >= self.max_players:
            raise InvalidOperation('Game already full')
        self.players[id_] = Player(lives=self.starting_lives)

    def remove_player(self, id_):
        """Remove a player and their data."""
        try:
            if id_ == self.dealer and len(self.players) != 1:
                keys = list(self.players.keys())
                idx = keys.index(id_)
                self.dealer = keys[idx - 1]
            del self.players[id_]
        except (KeyError, ValueError):
            raise ValueError('Player not found')
        # XXX Do not trigger a game end, because this might be called
        # when finishing the game. Not sure yet

    def next_player(self, id_, num=1):
        """Return the id of the next player(s), using the players dict
        as a circular dictionary.
        If the player is alone, id_ is returned.
        The method uses list.index, therefore if id_ is not a valid key
        of self.players, a ValueError is raised. As an exception, None
        can be used to retreive the first num player(s).
        """
        if num <= 0 or num > len(self.players):
            raise ValueError('Invalid number of players to fetch')
        keys = list(self.players.keys())
        if id_ is None:
            idx = 0
        else:
            idx = (keys.index(id_) + 1) % len(keys)
        if num == 1:
            return keys[idx]
        elif (idx + num) >= len(keys):
            keys = keys + keys
        return tuple(keys[idx: idx + num])

    def start(self, starting_cards=STARTING_CARDS):
        """Start the game."""
        if self.game_status >= GameStatus.PLAYING:
            raise InvalidOperation('Game already started')
        if len(self.players) < 2:
            raise InvalidOperation(
                'Cannot start a game with less than two players')

        self.starting_cards = starting_cards
        self.game_status = GameStatus.PLAYING
        try:
            self.next_hand(starting_cards)
        except:  # noqa: E722
            # We do not want to allow starting the game twice or more,
            # so we check the GameStatus. This is the way we though to
            # use to revert a failed start.
            self.game_status = GameStatus.WAIT_START
            raise

    def next_hand(self, num_cards=None):
        """Reset the status of the hand (cards, bets, etc) to start the
        next one.
        """
        if (self.game_status != GameStatus.PLAYING
                or self.hand_status != HandStatus.ENDED):
            raise InvalidOperation('Cannot start a hand right now')

        if isinstance(num_cards, int) and num_cards > 0:
            self.num_cards = num_cards
        elif self.num_cards <= 1:
            self.num_cards = self.starting_cards
        else:
            self.num_cards = self.num_cards - 1

        cards = self.card_class.make_cards(len(self.players), self.num_cards)
        i = 0
        for id_, pl in self.players.items():
            pl.reset(cards[i])
            i += 1
        self.bet_sum = 0
        self.reset_taking()

        self.dealer, self.turn = self.next_player(self.dealer, 2)
        self.began = self.turn
        self.hand_status = HandStatus.BET
        self.give_cards()
        self.ask_bet(self.turn)

    def bet(self, player, bet):
        """Place a bet, for a certain player.
        The function checks also if it is this player's turn.
        """
        if self.hand_status != HandStatus.BET:
            raise InvalidOperation('Bets are not open')
        if player != self.turn:
            raise InvalidOperation('Wrong turn')

        if not isinstance(bet, int) or bet < 0:
            raise ValueError('The bet must be a non-negative integer')
        if player == self.dealer:
            if self.bet_sum + bet == self.num_cards:
                raise ValueError('Invalid bet, for the dealer')

        self.players[player].bet = bet
        self.bet_sum += bet
        self.broadcast_bet(player, bet)
        self.forward_game()

    def move(self, player, card, high=False):
        """Perform a move for a certain player.
        The function checks also if it is this player's turn.
        """
        if self.hand_status != HandStatus.MOVE:
            raise InvalidOperation('You cannot move, right now')

        if player != self.turn:
            raise InvalidOperation('Wrong turn')
        try:
            cards = self.players[player].cards
        except KeyError:
            # Actually should never happen, if turn is valid
            raise ValueError('Invalid player')
        try:
            idx = cards.index(card)
        except ValueError:
            raise ValueError('The player does not have that card!')

        if high:
            card.set_high()
        # Remove only after it has been set high, because we let it
        # check if the card can be the highest/lowest, and it might
        # raise an exception.
        cards.pop(idx)
        if card > self.max or self.taking is None:
            self.max = card
            self.taking = player
            taking = True
        else:
            taking = False
        self.broadcast_move(player, card, taking)
        self.forward_game()

    def forward_game(self):
        """Method to call at the end of an action (i.e. bet or move) to
        make the game go into the next state.
        """
        next_pl = self.next_player(self.turn)
        if (self.hand_status == HandStatus.MOVE and next_pl == self.began):
            self.players[self.taking].took += 1
            self.turn = self.taking
            self.began = self.turn
            self.reset_taking()
            if not self.players[self.turn].cards:
                self.hand_status = HandStatus.ENDED
        else:
            if self.hand_status == HandStatus.BET and self.turn == self.dealer:
                self.hand_status = HandStatus.MOVE
            self.turn = next_pl

        if self.hand_status == HandStatus.BET:
            self.ask_bet(self.turn)
        elif self.hand_status == HandStatus.MOVE:
            cards = self.players[self.turn].cards
            if (len(cards) == 1 and not cards[0].eligible_high()
                    and self.enable_automove):
                self.move(self.turn, cards[0])
            else:
                self.ask_move(self.turn)
        elif self.hand_status == HandStatus.ENDED:
            self.finalize_hand()

    def finalize_hand(self):
        """Finalize a hand, and start another one, or end the game."""
        if self.hand_status != HandStatus.ENDED:
            raise InvalidOperation('Cannot finalize a hand that has not ended')

        items = list(self.players.items())
        dead = []
        for id_, pl in items:
            pl.reset()
            if pl.dead():
                dead.append(id_)
                del self.players[id_]

        if len(self.players) <= 1:
            self.game_status = GameStatus.ENDED
            try:
                maybe_winner = next(iter(self.players))
            except StopIteration:
                maybe_winner = None
            self.game_ended(maybe_winner)
        else:
            self.hand_ended(dead)
            self.next_hand()

    def reset_taking(self):
        """Reset the taking status.
        Implementations can override this method to also broadcast it
        to players."""
        self.taking = None
        self.max = None

    def give_cards(self):
        """Tell users their cards.
        Implementation can access cards through self.players and they
        should be careful of the special case for self.num_cards == 1.
        """
        pass

    def ask_bet(self, player):
        """Ask a player to bet."""
        pass

    def broadcast_bet(self, player, bet):
        """Broadcast a successful bet."""
        pass

    def ask_move(self, player):
        """Ask a player their move."""
        pass

    def broadcast_move(self, player, card, taking=False):
        """Broadcast a successful move.
        Implementation should be careful with the highest/lowest card.
        """
        pass

    def hand_ended(self, dead_players):
        """Broadcast that the hand ended and the player that took."""
        pass

    def game_ended(self, player):
        """Broadcast that the game has ended.
        player can also be None, if nobody won.
        """
        pass
