/* global io, alert */
'use strict'

/* Use a block to have $ inside with let, but be able to export functions,
which would not be possible with scope isolator */
{
  const JOIN_FRAGMENT = '#join-'
  const SUITES = ['❓', '🧨', '⚔️', '🏆', '💰']
  const HIGH_LOW_CARD = [1, 4]
  const GIVECARD_TIMEOUT = 500
  const DROPCARDS_TIMEOUT = 4000

  // Generic helpers and socket.io setup

  const $ = (sel, parent) => {
    if (parent === undefined) {
      parent = document
    }
    return parent.querySelector(sel)
  }

  const $$ = (sel, parent) => {
    if (parent === undefined) {
      parent = document
    }
    return parent.querySelectorAll(sel)
  }

  const showArea = (newClass) => {
    $('#wrapper').className = newClass
  }

  const deleteChildren = (node) => {
    while (node.hasChildNodes()) {
      node.removeChild(node.firstChild)
    }
  }

  const cloneTemplateNode = (templateSel, contentSel) => {
    var tpl = $(contentSel, $(templateSel).content)
    return document.importNode(tpl, true)
  }

  const log = text => {
    var li = document.createElement('li')
    li.textContent = text
    var logs = $('#logs')
    logs.insertBefore(li, logs.firstChild)
  }

  const clearLogs = () => {
    deleteChildren($('#logs'))
  }

  function getPlayerElement (player) {
    return $('[data-player-id="' + player + '"]')
  }

  var socket = io({
    transports: ['websocket']
  })

  socket.on('disconnect', () => {
    console.log('Server disconnected')
  })

  socket.on('exception', (message) => {
    alert('The server reported an error: ' + message)
  })

  socket.on('leave_game', () => {
    console.log('The server kicked you out.')
    showArea('home')
  })

  // Homepage: came creation and joining

  $('#create-form').onsubmit = () => {
    socket.emit('create', $('#create-user').value.trim(),
      $('#create-players').value.trim())
    return false
  }

  $('#join-form').onsubmit = () => {
    socket.emit('join', $('#join-id').value.trim(),
      $('#join-user').value.trim())
    return false
  }

  function checkHash () {
    var hash = window.location.hash
    if (hash.startsWith(JOIN_FRAGMENT)) {
      $('#join-id').value = hash.substr(JOIN_FRAGMENT.length)
      showArea('home')
    } else if (hash === '#home') {
      showArea('home')
    }
  }
  checkHash()
  window.onhashchange = checkHash

  socket.on('joined', (gameId, owner) => {
    console.info('Entering a game', gameId, owner)

    showArea('lobby')
    var cl = $('#lobby').classList
    if (owner) {
      cl.add('owned')
    } else {
      cl.remove('owned')
    }

    updateMatchUrl(gameId)
    $('#lobby-gid').textContent = gameId
    window.location.href = '#game-' + gameId
  })

  // Lobby
  var players = new Map()

  function updateMatchUrl (gameId) {
    var url = window.location.href
    var hash = url.indexOf('#')
    if (hash !== -1) {
      url = url.substr(0, hash)
    }
    url += JOIN_FRAGMENT + gameId
    $('#url-text').value = url
  }

  function createPlayerList () {
    var list = $('#player-list')
    deleteChildren(list)
    players.forEach((pl, id) => {
      var li = document.createElement('li')
      li.textContent = pl.name
      list.appendChild(li)
    })
  }

  socket.on('players', pl => {
    console.log('Received players', pl)
    players.clear()
    pl.forEach(player => {
      players.set(player[0], {
        name: player[1],
        lives: 5,
        taken: 0,
        bet: -1
      })
    })
    createPlayerList()

    var button = $('#start')
    if (pl.length > 1) {
      button.removeAttribute('disabled')
    } else {
      button.setAttribute('disabled', 'disabled')
    }
  })

  if (window.navigator.share) {
    $('body').classList.add('has-share')
  }

  $('#start').onclick = () => {
    socket.emit('start')
  }

  $('#url-copy').onclick = () => {
    $('#url-text').select()
    document.execCommand('copy')
  }

  $('#url-share').onclick = () => {
    navigator.share({
      title: 'Bisca',
      text: 'Join this Bisca match!',
      url: $('#url-text').value
    })
  }

  // Game

  // Workaround to prevent moving class to be removed
  var movingRequests = 0

  function createCard (card, highLow) {
    var cardElem = cloneTemplateNode('#card-template', '.card')
    if (card === undefined) {
      card = ['?', 0]
    }
    $('.card-number', cardElem).textContent = card[0]
    if ((highLow || highLow === undefined) &&
        card[0] === HIGH_LOW_CARD[0] && card[1] === HIGH_LOW_CARD[1]) {
      $('.card-number', cardElem).textContent += card[2] ? 'H' : 'L'
    }
    $('.card-suite', cardElem).textContent = SUITES[card[1]]
    return cardElem
  }

  function createYourCard (parent, card) {
    var cardElem = createCard(card, false)
    if (card) {
      cardElem.setAttribute('data-card-number', card[0])
      cardElem.setAttribute('data-card-suit', card[1])
    } else {
      cardElem.setAttribute('data-card-unknown', '')
    }
    if (card && card[0] === HIGH_LOW_CARD[0] && card[1] === HIGH_LOW_CARD[1]) {
      var highLow = cloneTemplateNode('#high-low-template', '.high-low')
      $$('button', highLow).forEach(button => {
        button.onclick = highLowClick
      })
      cardElem.appendChild(highLow)
      cardElem.onclick = () => {
        if (!$('#game').classList.contains('moving')) {
          return
        }
        resetHighLow()
        highLow.classList.add('active')
        return false
      }
    } else {
      cardElem.onclick = moveCard
    }
    parent.appendChild(cardElem)
    return parent
  }

  function moveCard (event) {
    if (!$('#game').classList.contains('moving')) {
      return false
    }

    resetHighLow()
    realMoveCard(event.currentTarget)
    return false
  }

  function highLowClick (event) {
    var button = event.currentTarget
    var high
    if (button.classList.contains('set-high')) {
      high = true
    } else {
      high = false
    }

    var card = button
    while (!card.classList.contains('card') && card.parentNode) {
      card = card.parentNode
    }
    realMoveCard(card, high)

    return false
  }

  function realMoveCard (card, highLow) {
    var unknown = card.hasAttribute('data-card-unknown')
    if (!unknown) {
      var number = card.getAttribute('data-card-number')
      var suit = card.getAttribute('data-card-suit')
    }

    var callback = approved => {
      if (approved) {
        card.parentNode.removeChild(card)
        movingRequests--
        if (!movingRequests) {
          $('#game').classList.remove('moving')
        }
      }
    }

    if (unknown) {
      // Will work, unless it was high/low card
      socket.emit('move_unknown', callback)
    } else if (highLow !== undefined) {
      socket.emit('move', number, suit, highLow, callback)
    } else {
      socket.emit('move', number, suit, callback)
    }
  }

  function resetHighLow () {
    var maybeHighLow = $('.high-low.active')
    if (maybeHighLow) {
      maybeHighLow.classList.remove('active')
    }
  }

  function setPlayerCard (player, card, highLow) {
    var plElem = getPlayerElement(player)
    var placeholder = $('.card-placeholder', plElem)
    if (!placeholder) {
      placeholder = $('.card', plElem)
    }
    plElem.replaceChild(createCard(card, highLow), placeholder)
  }

  $('#bet-form').onsubmit = () => {
    socket.emit('bet', $('#bet-value').value, approved => {
      if (approved) {
        $('#game').classList.remove('betting')
      }
    })
    return false
  }

  function updatePlayerInfo (player) {
    if (player === undefined) {
      players.forEach((player, id) => {
        updatePlayerInfo(id)
      })
      return
    }

    var plElem = getPlayerElement(player)
    if (!plElem || !players.has(player)) {
      return
    }
    var plInfo = players.get(player)
    $('.player-lives', plElem).textContent = plInfo.lives
    $('.player-taken', plElem).textContent = plInfo.taken
    if (plInfo.bet >= 0) {
      plElem.classList.add('has-bet')
      $('.player-bet-value', plElem).textContent = plInfo.bet
    } else {
      plElem.classList.remove('has-bet')
    }
  }

  function createPlayer (player) {
    if (getPlayerElement(player)) {
      return
    }
    var newPlayerEl = cloneTemplateNode('#player-template', '.player')
    newPlayerEl.setAttribute('data-player-id', player)
    $('.player-name', newPlayerEl).textContent = players.get(player).name
    $('#game-players').appendChild(newPlayerEl)
    updatePlayerInfo(player)
  }

  function deletePlayer (player) {
    var plElem = getPlayerElement(player)
    if (plElem) {
      plElem.parentNode.removeChild(plElem)
    }
    players.delete(player)
  }

  function createPlayers () {
    var container = $('#game-players')
    deleteChildren(container)
    players.forEach((player, id) => {
      createPlayer(id)
    })
  }

  function setPlayerStatus (player, status) {
    var old = $('.player.currently-' + status)
    if (old) {
      old.classList.remove('currently-' + status)
    }
    var plElem = getPlayerElement(player)
    if (plElem) {
      plElem.classList.add('currently-' + status)
    }
    console.log('Setting status', player, status, plElem)
  }

  function resetPlayerStatus (resetTaking) {
    console.log('Resetting!')
    setPlayerStatus('', 'betting')
    setPlayerStatus('', 'moving')
    if (resetTaking) {
      setPlayerStatus('', 'taking')
    }
  }

  // Game socket events

  let cardsCleaned = true
  let distributingCards = 0

  function clearCards () {
    if (cardsCleaned) {
      return
    }
    $$('#game-players .card').forEach(card => {
      var placeholder = document.createElement('div')
      placeholder.className = 'card-placeholder'
      card.parentNode.replaceChild(placeholder, card)
    })
    cardsCleaned = true
  }

  socket.on('game_started', () => {
    console.debug('Game started!')
    showArea('game')
    createPlayers()
    clearLogs()
  })

  socket.on('new_lap', (took) => {
    console.debug('New lap beginning', took)
    if (players.has(took)) {
      took = players.get(took)
      took.taken += 1
      log(took.name + ' took a lap')
    }

    updatePlayerInfo()
    resetPlayerStatus(true)
    cardsCleaned = false
    setTimeout(clearCards, DROPCARDS_TIMEOUT)
  })

  socket.on('your_cards', (cards, dealer) => {
    console.debug('Cards arrived', cards)
    var container = $('#your-cards ul')
    deleteChildren(container)
    var callback
    callback = () => {
      container.appendChild(createYourCard(document.createElement('li'),
        cards.shift()))
      distributingCards = cards.length
      if (cards.length) {
        setTimeout(callback, GIVECARD_TIMEOUT)
      }
    }
    callback()
    $('#dealer').textContent = players.get(dealer).name
  })

  socket.on('table_cards', (cards, dealer) => {
    console.debug('Table cards received', cards)
    clearCards()
    Object.keys(cards).forEach(id => {
      setPlayerCard(id, cards[id], false)
    })
    $('#your-cards ul').appendChild(createYourCard(
      document.createElement('li')))
    $('#dealer').textContent = players.get(dealer).name
  })

  socket.on('please_bet', () => {
    setTimeout(() => {
      $('#game').classList.add('betting')
    }, distributingCards * GIVECARD_TIMEOUT + 100)
  })

  socket.on('please_move', () => {
    console.debug('The server is asking for a move')
    $('#game').classList.add('moving')
    movingRequests++
  })

  socket.on('bet', (player, bet) => {
    console.debug('Received bet', player, bet)
    players.get(player).bet = bet
    updatePlayerInfo(player)
  })

  socket.on('move', (player, card, taking) => {
    console.debug('New move', player, card, taking)
    clearCards()
    setPlayerCard(player, card)

    if (taking) {
      console.log('Set taking!')
      setPlayerStatus(player, 'taking')
    }
  })

  socket.on('game_status', (status, player) => {
    resetPlayerStatus()
    console.debug('Change of game status', status, player)
    setPlayerStatus(player, status)
  })

  socket.on('hand_ended', (deadPlayers, lives) => {
    console.debug('Hand ended', deadPlayers, lives)
    clearLogs()
    deadPlayers.forEach(id => {
      log(players.get(id).name + ' has lost all their lives')
      deletePlayer(id)
    })
    players.forEach((player, id) => {
      player.lives = lives[id]
      player.taken = 0
      player.bet = -1
    })
    updatePlayerInfo()
  })

  socket.on('game_ended', maybeWinner => {
    console.debug('Game ended', maybeWinner)
    showArea('game-over')
    if (maybeWinner) {
      $('#game-over').className = 'somebody'
      $('#winner-placeholder').textContent = players.get(maybeWinner).name
    } else {
      $('#game-over').className = 'nobody'
    }
  })
}
