import uuid

import socketio

from game import Game, InvalidOperation


class SioGame(Game):

    def __init__(self, sio, owner, owner_name, max_players=None):
        if max_players is not None:
            super().__init__(max_players=max_players)
        else:
            super().__init__()

        self.sio = sio
        self.id = str(uuid.uuid4())
        self.room = 'game_' + str(self.id)
        self.add_player(owner, owner_name)
        self.owner = owner

    def add_player(self, sid, name):
        super().add_player(sid)
        self.players[sid].name = name
        self.sio.enter_room(sid, self.room)
        self.broadcast_players()

    def remove_player(self, sid):
        super().remove_player(sid)
        self.sio.leave_room(sid, self.room)
        self.sio.emit('leave_game', (), room=sid)
        self.broadcast_players()

    def start(self):
        self.sio.emit('game_started')
        super().start()

    def reset_taking(self):
        took = self.taking if hasattr(self, 'taking') else None
        super().reset_taking()
        self.sio.emit('new_lap', took, room=self.room)

    def give_cards(self):
        if self.num_cards == 1:
            cards = {sid: pl.cards[0].serialize()
                     for sid, pl in self.players.items()}
            for sid in cards:
                to_send = dict(cards)
                del to_send[sid]
                self.sio.emit('table_cards', (to_send, self.dealer), room=sid)
        else:
            for sid in self.players.keys():
                self.give_player_cards(sid)

    def give_player_cards(self, sid):
        cards = [c.serialize() for c in self.players[sid].cards]
        self.sio.emit('your_cards', (cards, self.dealer), room=sid)

    def ask_bet(self, player):
        self.sio.emit('please_bet', (), room=player)
        self.sio.emit('game_status', ('betting', player), room=self.room)

    def broadcast_bet(self, player, bet):
        self.sio.emit('bet', (player, bet), room=self.room)

    def ask_move(self, player):
        self.sio.emit('please_move', (), room=player)
        self.sio.emit('game_status', ('moving', player), room=self.room)

    def broadcast_move(self, player, card, taking=False):
        self.sio.emit(
            'move', (player, card.serialize(), taking), room=self.room)

    def hand_ended(self, dead_players):
        lives = {sid: pl.lives for sid, pl in self.players.items()}
        self.sio.emit(
            'hand_ended', (dead_players, lives), room=self.room)

    def game_ended(self, player):
        self.sio.emit(
            'game_ended', player, room=self.room)
        for sid in self.players:
            self.sio.disassociate_player(sid)
        self.sio.forget_game(self.id)

    def broadcast_players(self):
        players = [[sid, pl.name] for sid, pl in self.players.items()]
        self.sio.emit('players', players, room=self.room)

    def move_unknown(self, player):
        """Move the card on the single card stage."""
        try:
            cards = self.players[player].cards
        except KeyError:
            raise ValueError('Invalid player')
        if len(cards) != 1:
            raise InvalidOperation('You have more than one card')
        if cards[0].eligible_high():
            return False
        self.move(player, cards[0])
        return True


class SioGameNamespace(socketio.Namespace):

    def __init__(self, namespace=None):
        super().__init__(namespace)
        # A map {game id: game object} (used to join games)
        self.games = {}
        # A map {sid: game object} (used for all events)
        self.player_games = {}

    def disassociate_player(self, sid):
        """Remove a player from the sid-game dictionary.
        This function is though to be called by the game itself,
        therefore it does not remove the player from the game.
        """
        if sid in self.player_games:
            del self.player_games[sid]

    def forget_game(self, gid):
        """Forget a game, because it has ended."""
        if gid in self.games:
            del self.games[gid]

    def send_exception(self, sid, message):
        """Report an exception to a player."""
        self.emit('exception', message, room=sid)

    def get_player_game(self, sid):
        """Return the game associated to a user, or send an error on
        the socket, if the user is not associated to any game.
        """
        game = self.player_games.get(sid, None)
        if game is None:
            self.send_exception(sid, 'You are not associated to any game.')
        return game

    def on_disconnect(self, sid):
        """Handle player disconnection."""
        if sid in self.player_games:
            try:
                self.player_games[sid].remove_player(sid)
            except ValueError:
                # Already deleted, so ignore the problem
                pass
            del self.player_games[sid]

    def on_create(self, sid, player_name, max_players=None):
        """Handle the create game event."""
        # TODO: Handle removal from other games

        if max_players is not None:
            try:
                max_players = int(max_players)
            except ValueError:
                self.send_exception(sid, 'Invalid number of max players.')
                return

        try:
            game = SioGame(self, sid, player_name, max_players)
            self.games[game.id] = game
            self.player_games[sid] = game
            self.emit('joined', (game.id, True), room=sid)
        except ValueError as e:
            self.send_exception(sid, str(e))

    def on_join(self, sid, game_id, player_name):
        """Handle the join game event."""
        # TODO: Handle removal from other games
        try:
            game = self.games[game_id]
            game.add_player(sid, player_name)
            self.player_games[sid] = game
            self.emit('joined', (game_id, False), room=sid)
        except KeyError:
            self.send_exception(sid, 'Game not found')
        except (ValueError, InvalidOperation) as e:
            self.send_exception(sid, str(e))

    def on_start(self, sid):
        """Handle the start game event."""
        game = self.get_player_game(sid)
        if game is None:
            return

        if game.owner == sid:
            try:
                game.start()
            except InvalidOperation as e:
                self.send_exception(
                    sid, 'You cannot start the game right now. ' + str(e))
        else:
            self.send_exception(
                sid, 'You must be the game owner to start it')

    def on_bet(self, sid, bet):
        """Handle the bet event."""
        try:
            bet = int(bet)
        except (ValueError, TypeError):
            self.send_exception(sid, 'Invalid bet')
            return
        game = self.get_player_game(sid)
        if game is None:
            return

        try:
            game.bet(sid, bet)
            return True
        except (ValueError, InvalidOperation) as e:
            self.send_exception(sid, str(e))

    def on_move(self, sid, number, suit, high=False):
        """Handle the move event."""
        game = self.get_player_game(sid)
        if game is None:
            return

        try:
            number = int(number)
            suit = int(suit)
            high = bool(high)
        except ValueError:
            self.send_exception(sid, 'Wrong data')
            return

        try:
            card = game.card_class(number, suit)
            game.move(sid, card, high)
            return True
        except (ValueError, InvalidOperation) as e:
            self.send_exception(sid, str(e))

    def on_move_unknown(self, sid):
        """Handle the move unknown card event."""
        game = self.get_player_game(sid)
        if game is None:
            return
        try:
            if game.move_unknown(sid):
                return True
            else:
                # More than one card is an error, whereas the function
                # returns false only when the card is the
                # highest/lowest
                game.give_player_cards(sid)
        except (ValueError, InvalidOperation) as e:
            self.send_exception(sid, str(e))
