import random

import pytest

from game import Game, InvalidOperation, GameStatus, HandStatus, Card, CardSuit


def test_coins_ace():
    low = Card(1, CardSuit.COINS)
    assert low == low
    assert not low > low
    assert not low < low
    assert low > None
    assert low < Card(1, CardSuit.STICKS)
    assert Card(1, CardSuit.STICKS) > low

    high = Card(1, CardSuit.COINS, True)
    assert high == high
    assert not high > high
    assert not high < high
    # High is a state, not an identity... Something like that
    assert low == high
    # Although it affects comparisons
    assert high > low
    assert low < high
    assert high > Card(10, CardSuit.COINS)
    assert Card(10, CardSuit.COINS) < high


def test_none_player():
    g = Game()
    with pytest.raises(ValueError):
        g.add_player(None)


def test_unique_players():
    g = Game()
    g.add_player('Piero')
    with pytest.raises(ValueError):
        g.add_player('Piero')


def test_max_players():
    with pytest.raises(ValueError):
        g = Game(max_players=1000)
    with pytest.raises(ValueError):
        g = Game(max_players=1)
    with pytest.raises(ValueError):
        g = Game(max_players=0)
    with pytest.raises(ValueError):
        g = Game(max_players=-1)

    g = Game(max_players=2)
    g.add_player('Piero')
    g.add_player('Nico')
    with pytest.raises(InvalidOperation):
        g.add_player('Bosca')


def test_remove_player():
    g = Game()
    g.add_player('Piero')
    g.remove_player('Piero')
    assert len(g.players) == 0
    with pytest.raises(ValueError):
        g.remove_player('Piero')


def test_next_player():
    g = Game()

    g.add_player('Piero')
    assert g.next_player(None) == 'Piero'
    assert g.next_player('Piero') == 'Piero'
    with pytest.raises(ValueError):
        assert g.next_player('Piero', 2)

    g.add_player('Nico')
    assert g.next_player('Piero') == 'Nico'
    assert g.next_player('Nico') == 'Piero'
    assert g.next_player(None, 2) == ('Piero', 'Nico')


def test_start():
    g = Game()
    g.add_player('Piero')
    with pytest.raises(InvalidOperation):
        # Cannot start with only one player
        g.start()
    g.add_player('Nico')
    with pytest.raises(ValueError):
        # Not enough cards
        g.start(25)

    num_cards = 5
    g.start(num_cards)
    for name, pl in g.players.items():
        assert len(pl.cards) == num_cards
    assert g.game_status == GameStatus.PLAYING
    assert g.dealer == 'Piero'
    assert g.turn == 'Nico'


def test_bet():
    g = Game()
    g.add_player('Piero')
    g.add_player('Nico')

    with pytest.raises(InvalidOperation):
        # Betting before the start
        g.bet('Nico', 2)
    g.start(5)

    with pytest.raises(ValueError):
        # Non integer bet
        g.bet('Nico', 0.5)
    with pytest.raises(ValueError):
        # Negative bet
        g.bet('Nico', -1)

    with pytest.raises(InvalidOperation):
        # Wrong turn
        g.bet('Piero', 2)
    g.bet('Nico', 2)
    with pytest.raises(ValueError):
        # Sum of cards of the dealer
        g.bet('Piero', 3)
    g.bet('Piero', 2)

    assert g.hand_status == HandStatus.MOVE
    assert g.turn == 'Nico'


def test_move():
    # Piero: 3S, 5C; Nico: 9B, 4S
    random.seed(1234)
    g = Game(starting_lives=2)
    g.add_player('Piero')
    g.add_player('Nico')
    g.start(starting_cards=2)

    with pytest.raises(InvalidOperation):
        # Move before betting
        g.move('Nico', Card(9, CardSuit.STICKS))
    g.bet('Nico', 1)
    g.bet('Piero', 2)

    with pytest.raises(InvalidOperation):
        # Invalid turn/player
        g.move('Nico2', Card(9, CardSuit.STICKS))
    with pytest.raises(ValueError):
        # Move an unowned card
        g.move('Nico', Card(10, CardSuit.STICKS))

    g.move('Nico', Card(9, CardSuit.STICKS))
    assert g.taking == 'Nico'
    assert g.max == Card(9, CardSuit.STICKS)
    assert g.turn == 'Piero'
    g.move('Piero', Card(3, CardSuit.SWORDS))
    assert g.players['Piero'].took == 1
    assert g.players['Nico'].took == 0
    assert g.players['Piero'].lives == 2
    assert g.players['Nico'].lives == 2

    assert g.turn == 'Piero'
    g.move('Piero', Card(5, CardSuit.CUPS))
    g.move('Nico', Card(4, CardSuit.SWORDS))
    assert g.players['Piero'].lives == 2
    assert g.players['Nico'].lives == 1

    # Next hand
    # Piero: 9S, Nico: 3D
    assert g.dealer == 'Nico'
    g.bet('Piero', 0)
    g.bet('Nico', 0)
    g.move('Piero', Card(9, CardSuit.SWORDS))
    g.move('Nico', Card(3, CardSuit.COINS))
    assert g.players['Piero'].lives == 2
    # When players die, they are removed from the dict
    assert 'Nico' not in g.players
    assert g.game_status == GameStatus.ENDED

# TODO: Test with user quitting
# TODO: Test with coin ace
# TODO: Test events
